# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=line-too-long,too-many-lines

import os
import time

from azure.cli.core.util import get_file_json
from azure.core.exceptions import HttpResponseError
from azure.mgmt.servicefabric.models import (ApplicationTypeResource,
                                             ApplicationTypeVersionResource,
                                             ApplicationResource,
                                             ApplicationUpgradePolicy,
                                             ArmRollingUpgradeMonitoringPolicy,
                                             ArmApplicationHealthPolicy,
                                             ArmServiceTypeHealthPolicy)
from azure.cli.command_modules.servicefabric._arm_deployment_utils import validate_and_deploy_arm_template

from knack.log import get_logger

logger = get_logger(__name__)


def create_app(client,
               resource_group_name,
               cluster_name,
               application_type_name,
               application_type_version,
               application_name,
               package_url=None,
               application_parameters=None,
               minimum_nodes=None,
               maximum_nodes=None):
    if package_url is not None:
        create_app_type_version(client, resource_group_name, cluster_name, application_type_name, application_type_version, package_url)

    try:
        apps = client.applications.list(resource_group_name, cluster_name)
        for app in apps.value:
            if app.name.lower() == application_name.lower():
                logger.info("Application '%s' already exists", application_name)
                return app

        appResource = ApplicationResource(type_name=application_type_name,
                                          type_version=application_type_version,
                                          minimum_nodes=minimum_nodes,
                                          maximum_nodes=maximum_nodes,
                                          parameters=application_parameters)
        appResource.name = application_name
        app = client.applications.begin_create_or_update(resource_group_name, cluster_name, application_name, appResource).result()
        return app
    except HttpResponseError as ex:
        logger.error("HttpResponseError: %s", ex)
        raise


def update_app(client,
               resource_group_name,
               cluster_name,
               application_name,
               application_type_version=None,
               application_parameters=None,
               minimum_nodes=None,
               maximum_nodes=None,
               force_restart=False,
               upgrade_replica_set_check_timeout=None,
               failure_action=None,
               health_check_retry_timeout=None,
               health_check_wait_duration=None,
               health_check_stable_duration=None,
               upgrade_domain_timeout=None,
               upgrade_timeout=None,
               consider_warning_as_error=False,
               default_service_type_max_percent_unhealthy_partitions_per_service=None,
               default_service_type_max_percent_unhealthy_replicas_per_partition=None,
               default_service_type_max_percent_unhealthy_services=None,
               max_percent_unhealthy_deployed_applications=None,
               service_type_health_policy_map=None):
    try:
        currentApp = client.applications.get(resource_group_name, cluster_name, application_name)
        appResource = currentApp
        # TODO: change to patch once the fix is deployed in the rp
        # appResourceUpdate: ApplicationResourceUpdate = ApplicationResourceUpdate()

        if application_type_version:
            appResource.type_version = application_type_version
        if application_parameters:
            appResource.parameters.update(application_parameters)
        if minimum_nodes is not None:
            appResource.minimum_nodes = minimum_nodes
        if maximum_nodes is not None:
            appResource.maximum_nodes = maximum_nodes

        appResource.upgrade_policy = _set_uprade_policy(currentApp.upgrade_policy,
                                                        force_restart,
                                                        upgrade_replica_set_check_timeout,
                                                        failure_action,
                                                        health_check_retry_timeout,
                                                        health_check_wait_duration,
                                                        health_check_stable_duration,
                                                        upgrade_domain_timeout,
                                                        upgrade_timeout,
                                                        consider_warning_as_error,
                                                        default_service_type_max_percent_unhealthy_partitions_per_service,
                                                        default_service_type_max_percent_unhealthy_replicas_per_partition,
                                                        default_service_type_max_percent_unhealthy_services,
                                                        max_percent_unhealthy_deployed_applications,
                                                        service_type_health_policy_map)

        # TODO: change to patch once the fix is deployed in the rp
        # client.applications.update(resource_group_name, cluster_name, application_name, appResourceUpdate)
        return client.applications.begin_create_or_update(resource_group_name, cluster_name, application_name, appResource).result()
    except HttpResponseError as ex:
        logger.error("HttpResponseError: %s", ex)
        raise


def create_app_type(client, resource_group_name, cluster_name, application_type_name):
    try:
        appTypes = client.application_types.list(resource_group_name, cluster_name)
        for appType in appTypes.value:
            if appType.name.lower() == application_type_name.lower():
                logger.info("Application type '%s' already exists", application_type_name)
                return appType

        appTypeResource = ApplicationTypeResource()
        logger.info("Creating application type '%s'", application_type_name)
        return client.application_types.create_or_update(resource_group_name, cluster_name, application_type_name, appTypeResource)
    except HttpResponseError as ex:
        logger.error("HttpResponseError: %s", ex)
        raise


def create_app_type_version(client,
                            resource_group_name,
                            cluster_name,
                            application_type_name,
                            version,
                            package_url):
    create_app_type(client, resource_group_name, cluster_name, application_type_name)
    try:
        appTypeVerions = client.application_type_versions.list(resource_group_name, cluster_name, application_type_name)
        for appTypeVerion in appTypeVerions.value:
            if appTypeVerion.name.lower() == version.lower():
                logger.info("Application type version '%s' already exists", version)
                return appTypeVerion

        appTypeVersionResource = ApplicationTypeVersionResource(app_package_url=package_url)
        logger.info("Creating application type version %s:%s", application_type_name, version)
        return client.application_type_versions.begin_create_or_update(resource_group_name,
                                                                       cluster_name,
                                                                       application_type_name,
                                                                       version,
                                                                       appTypeVersionResource).result()
    except HttpResponseError as ex:
        logger.error("HttpResponseError: %s", ex)
        raise


def create_service(cmd,
                   client,
                   resource_group_name,
                   cluster_name,
                   application_name,
                   service_name,
                   service_type,
                   state,
                   instance_count=None,
                   target_replica_set_size=None,
                   min_replica_set_size=None,
                   default_move_cost=None,
                   partition_scheme='singleton'):
    parameter_file, template_file = _get_template_file_and_parameters_file()
    template = get_file_json(template_file)
    parameters = get_file_json(parameter_file)['parameters']

    # set params
    _set_parameters(parameters, "clusterName", cluster_name)
    _set_parameters(parameters, "applicationName", application_name)
    _set_parameters(parameters, "serviceName", service_name)

    _set_service_parameters(template, parameters, "serviceTypeName", service_type, "string")

    if partition_scheme == 'singleton':
        _set_service_parameters(template, parameters, "partitionDescription", {"partitionScheme": "Singleton"}, "object")
    elif partition_scheme == 'uniformInt64':
        _set_service_parameters(template, parameters, "partitionDescription", {"partitionScheme": "UniformInt64Range"}, "object")
    elif partition_scheme == 'named':
        _set_service_parameters(template, parameters, "partitionDescription", {"partitionScheme": "Named"}, "object")

    if state == 'stateless':
        _set_service_parameters(template, parameters, "instanceCount", int(instance_count), "int")
    else:
        _set_service_parameters(template, parameters, "targetReplicaSetSize", int(target_replica_set_size), "int")
        _set_service_parameters(template, parameters, "minReplicaSetSize", int(min_replica_set_size), "int")

    if default_move_cost:
        _set_service_parameters(template, parameters, "defaultMoveCost", default_move_cost, "string")

    validate_and_deploy_arm_template(cmd, resource_group_name, template, parameters)

    return client.services.get(resource_group_name, cluster_name, application_name, service_name)


def _get_template_file_and_parameters_file():
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    template_parameter_folder = os.path.join('template', 'service')
    parameter_file = os.path.join(
        script_dir, template_parameter_folder, 'parameter.json')
    template_file = os.path.join(
        script_dir, template_parameter_folder, 'template.json')
    return parameter_file, template_file


def _set_service_parameters(template, parameters, name, value, param_type):
    tempalte_parameters = template['parameters']
    if name not in tempalte_parameters:
        tempalte_parameters[name] = {}
    tempalte_parameters[name]["type"] = param_type
    tempalte_resources_properties = template['resources'][0]['properties']
    tempalte_resources_properties[name] = "[parameters('{}')]".format(name)
    _set_parameters(parameters, name, value)


def _set_parameters(parameters, name, value):
    if name not in parameters:
        parameters[name] = {}
    parameters[name]["value"] = value


def _set_uprade_policy(current_upgrade_policy,
                       force_restart,
                       upgrade_replica_set_check_timeout,
                       failure_action,
                       health_check_retry_timeout,
                       health_check_wait_duration,
                       health_check_stable_duration,
                       upgrade_domain_timeout,
                       upgrade_timeout,
                       consider_warning_as_error,
                       default_service_type_max_percent_unhealthy_partitions_per_service,
                       default_service_type_max_percent_unhealthy_replicas_per_partition,
                       default_max_percent_service_type_unhealthy_services,
                       max_percent_unhealthy_deployed_applications,
                       service_type_health_policy_map):
    if current_upgrade_policy is None:
        current_upgrade_policy = ApplicationUpgradePolicy()

    if force_restart:
        current_upgrade_policy.force_restart = force_restart
    if upgrade_replica_set_check_timeout is not None:
        current_upgrade_policy.upgrade_replica_set_check_timeout = time.strftime('%H:%M:%S', time.gmtime(upgrade_replica_set_check_timeout))

    # RollingUpgradeMonitoringPolicy
    if current_upgrade_policy.rolling_upgrade_monitoring_policy is None:
        # initialize with defaults
        current_upgrade_policy.rolling_upgrade_monitoring_policy \
            = ArmRollingUpgradeMonitoringPolicy(failure_action='Manual',
                                                health_check_stable_duration=time.strftime('%H:%M:%S', time.gmtime(120)),
                                                health_check_retry_timeout=time.strftime('%H:%M:%S', time.gmtime(600)),
                                                health_check_wait_duration=time.strftime('%H:%M:%S', time.gmtime(0)),
                                                upgrade_timeout=time.strftime('%H:%M:%S', time.gmtime(86399)),
                                                upgrade_domain_timeout=time.strftime('%H:%M:%S', time.gmtime(86399)))

    if failure_action:
        current_upgrade_policy.rolling_upgrade_monitoring_policy.failure_action = failure_action
    if health_check_stable_duration is not None:
        current_upgrade_policy.rolling_upgrade_monitoring_policy.health_check_stable_duration = time.strftime('%H:%M:%S', time.gmtime(health_check_stable_duration))
    if health_check_retry_timeout is not None:
        current_upgrade_policy.rolling_upgrade_monitoring_policy.health_check_retry_timeout = time.strftime('%H:%M:%S', time.gmtime(health_check_retry_timeout))
    if health_check_wait_duration is not None:
        current_upgrade_policy.rolling_upgrade_monitoring_policy.health_check_wait_duration = time.strftime('%H:%M:%S', time.gmtime(health_check_wait_duration))
    if upgrade_timeout is not None:
        current_upgrade_policy.rolling_upgrade_monitoring_policy.upgrade_timeout = time.strftime('%H:%M:%S', time.gmtime(upgrade_timeout))
    if upgrade_domain_timeout is not None:
        current_upgrade_policy.rolling_upgrade_monitoring_policy.upgrade_domain_timeout = time.strftime('%H:%M:%S', time.gmtime(upgrade_domain_timeout))

    # ApplicationHealthPolicy
    if current_upgrade_policy.application_health_policy is None:
        current_upgrade_policy.application_health_policy = ArmApplicationHealthPolicy()

    if consider_warning_as_error:
        current_upgrade_policy.application_health_policy.consider_warning_as_error = True

    if current_upgrade_policy.application_health_policy.default_service_type_health_policy is None:
        current_upgrade_policy.application_health_policy.default_service_type_health_policy = ArmServiceTypeHealthPolicy(
            max_percent_unhealthy_partitions_per_service=default_service_type_max_percent_unhealthy_partitions_per_service,
            max_percent_unhealthy_replicas_per_partition=default_service_type_max_percent_unhealthy_replicas_per_partition,
            max_percent_unhealthy_services=default_max_percent_service_type_unhealthy_services)
    else:
        if default_service_type_max_percent_unhealthy_partitions_per_service:
            current_upgrade_policy.application_health_policy.default_service_type_health_policy .max_percent_unhealthy_partitions_per_service \
                = default_service_type_max_percent_unhealthy_partitions_per_service
        if default_service_type_max_percent_unhealthy_replicas_per_partition:
            current_upgrade_policy.application_health_policy.default_service_type_health_policy.max_percent_unhealthy_replicas_per_partition \
                = default_service_type_max_percent_unhealthy_replicas_per_partition
        if default_max_percent_service_type_unhealthy_services:
            current_upgrade_policy.application_health_policy.default_service_type_health_policy.max_percent_unhealthy_partitions_per_service \
                = default_max_percent_service_type_unhealthy_services

    if max_percent_unhealthy_deployed_applications:
        current_upgrade_policy.ApplicationHealthPolicy.max_percent_unhealthy_deployed_applications \
            = max_percent_unhealthy_deployed_applications

    if service_type_health_policy_map:
        current_upgrade_policy.application_health_policy.service_type_health_policy_map = service_type_health_policy_map
    return current_upgrade_policy



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------


def get_cognitiveservices_management_client(cli_ctx, *_):
    from azure.cli.core.commands.client_factory import get_mgmt_service_client
    from azure.mgmt.cognitiveservices import CognitiveServicesManagementClient
    return get_mgmt_service_client(cli_ctx, CognitiveServicesManagementClient)


def cf_accounts(cli_ctx, *_):
    return get_cognitiveservices_management_client(cli_ctx).accounts


def cf_deleted_accounts(cli_ctx, *_):
    return get_cognitiveservices_management_client(cli_ctx).deleted_accounts


def cf_resource_skus(cli_ctx, *_):
    return get_cognitiveservices_management_client(cli_ctx).resource_skus



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------


import hashlib
import hmac


class KDF:
    @staticmethod
    def to_big_endian_32bits(value):
        result = bytearray()
        result.append((value & 0xFF000000) >> 24)
        result.append((value & 0x00FF0000) >> 16)
        result.append((value & 0x0000FF00) >> 8)
        result.append(value & 0x000000FF)
        return result

    @staticmethod
    def to_big_endian_64bits(value):
        result = bytearray()
        result.append((value & 0xFF00000000000000) >> 56)
        result.append((value & 0x00FF000000000000) >> 48)
        result.append((value & 0x0000FF0000000000) >> 40)
        result.append((value & 0x000000FF00000000) >> 32)
        result.append((value & 0x00000000FF000000) >> 24)
        result.append((value & 0x0000000000FF0000) >> 16)
        result.append((value & 0x000000000000FF00) >> 8)
        result.append(value & 0x00000000000000FF)
        return result

    @staticmethod
    def test_sp800_108():
        label = 'label'
        context = 'context'
        bit_length = 256
        hex_result = 'f0ca51f6308791404bf68b56024ee7c64d6c737716f81d47e1e68b5c4e399575'
        key = bytearray()
        key.extend([0x41] * 32)

        new_key = KDF.sp800_108(key, label, context, bit_length)
        hex_value = new_key.hex().replace('-', '')
        return hex_value.lower() == hex_result

    @staticmethod
    def sp800_108(key_in: bytearray, label: str, context: str, bit_length):
        """
        Note - initialize out to be the number of bytes of keying material that you need
        This implements SP 800-108 in counter mode, see section 5.1

        Fixed values:
            1. h - The length of the output of the PRF in bits, and
            2. r - The length of the binary representation of the counter i.

        Input: KI, Label, Context, and L.

        Process:
            1. n := ⎡L/h⎤.
            2. If n > 2^(r-1), then indicate an error and stop.
            3. result(0):= ∅.
            4. For i = 1 to n, do
                a. K(i) := PRF (KI, [i]2 || Label || 0x00 || Context || [L]2)
                b. result(i) := result(i-1) || K(i).

            5. Return: KO := the leftmost L bits of result(n).
        """

        if bit_length <= 0 or bit_length % 8 != 0:
            return None

        L = bit_length
        bytes_needed = bit_length // 8
        hMAC = hmac.HMAC(key_in, digestmod=hashlib.sha512)
        hash_bits = hMAC.digest_size
        n = L // hash_bits
        if L % hash_bits != 0:
            n += 1

        hmac_data_suffix = bytearray()
        hmac_data_suffix.extend(label.encode('UTF-8'))
        hmac_data_suffix.append(0)
        hmac_data_suffix.extend(context.encode('UTF-8'))
        hmac_data_suffix.extend(KDF.to_big_endian_32bits(bit_length))

        out_value = bytearray()
        for i in range(n):
            hmac_data = bytearray()
            hmac_data.extend(KDF.to_big_endian_32bits(i + 1))
            hmac_data.extend(hmac_data_suffix)
            hMAC.update(hmac_data)
            hash_value = hMAC.digest()

            if bytes_needed > len(hash_value):
                out_value.extend(hash_value)
                bytes_needed -= len(hash_value)
            else:
                out_value.extend(hash_value[len(out_value): len(out_value) + bytes_needed])
                return out_value

        return None


if __name__ == '__main__':
    print(KDF.test_sp800_108())



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from knack.help_files import helps

# pylint: disable=line-too-long

helps['logicapp'] = """
type: group
short-summary: Manage logic apps.
"""

helps['logicapp create'] = """
type: command
short-summary: Create a logic app.
long-summary: The logic app's name must be able to produce a unique FQDN as AppName.azurewebsites.net.
examples:
  - name: Create a basic logic app.
    text: >
        az logicapp create -g MyResourceGroup --subscription MySubscription -p MyPlan -n MyUniqueAppName -s MyStorageAccount
"""

helps['logicapp delete'] = """
type: command
short-summary: Delete a logic app.
examples:
  - name: Delete a logic app.
    text: az logicapp delete --name MyLogicapp --resource-group MyResourceGroup --subscription MySubscription
"""

helps['logicapp show'] = """
type: command
short-summary: Get the details of a logic app.
examples:
  - name: Get the details of a logic app.
    text: az logicapp show --name MyLogicapp --resource-group MyResourceGroup --subscription MySubscription
"""

helps['logicapp list'] = """
type: command
short-summary: List logic apps.
examples:
  - name: List default host name and state for all logic apps.
    text: >
        az logicapp list --query "[].{hostName: defaultHostName, state: state}"
  - name: List all running logic apps.
    text: >
        az logicapp list --query "[?state=='Running']"
"""



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import json
from collections import OrderedDict
from knack.log import get_logger

logger = get_logger(__name__)


def configstore_output_format(result):
    return _output_format(result, _configstore_format_group)


def configstore_identity_format(result):
    return _output_format(result, _configstore_identity_format_group)


def configstore_credential_format(result):
    return _output_format(result, _configstore_credential_format_group)


def keyvalue_entry_format(result):
    return _output_format(result, _keyvalue_entry_format_group)


def featureflag_entry_format(result):
    return _output_format(result, _featureflag_entry_format_group)


def featurefilter_entry_format(result):
    return _output_format(result, _featurefilter_entry_format_group)


def _output_format(result, format_group):
    if 'value' in result and isinstance(result['value'], list):
        result = result['value']
    obj_list = result if isinstance(result, list) else [result]
    return [format_group(item) for item in obj_list]


def _configstore_format_group(item):
    sku_value = _get_value(item, 'sku')

    try:
        sku = json.loads(sku_value.replace("\'", "\"")).get('name')
    except ValueError:
        logger.debug("No valid sku %s found", sku_value)
        sku = ""

    return OrderedDict([
        ('CREATION DATE', _format_datetime(_get_value(item, 'creationDate'))),
        ('ENDPOINT', _get_value(item, 'endpoint')),
        ('LOCATION', _get_value(item, 'location')),
        ('NAME', _get_value(item, 'name')),
        ('PROVISIONING STATE', _get_value(item, 'provisioningState')),
        ('RESOURCE GROUP', _get_value(item, 'resourceGroup')),
        ('SKU', sku)
    ])


def _configstore_identity_format_group(item):
    return OrderedDict([
        ('PRINCIPAL ID', _format_datetime(_get_value(item, 'principalId'))),
        ('TENANT ID', _get_value(item, 'tenantId')),
        ('TYPE', _get_value(item, 'type'))
    ])


def _configstore_credential_format_group(item):
    return OrderedDict([
        ('CONNECTION STRING', _get_value(item, 'connectionString')),
        ('ID', _get_value(item, 'id')),
        ('LAST MODIFIED', _format_datetime(_get_value(item, 'lastModified'))),
        ('NAME', _get_value(item, 'name')),
        ('READ ONLY', _get_value(item, 'readOnly')),
        ('VALUE', _get_value(item, 'value'))
    ])


def _keyvalue_entry_format_group(item):
    # CLI core converts KeyValue object field names to camelCase (eg: content_type becomes contentType)
    # But when customers specify field filters, we return a dict of requested fields instead of KeyValue object
    # In that case, field names are not converted to camelCase. We need to check for both content_type and contentType
    content_type = _get_value(item, 'contentType')
    content_type = content_type if content_type != ' ' else _get_value(item, 'content_type')

    last_modified = _get_value(item, 'lastModified')
    last_modified = last_modified if last_modified != ' ' else _get_value(item, 'last_modified')

    return OrderedDict([
        ('CONTENT TYPE', content_type),
        ('KEY', _get_value(item, 'key')),
        ('VALUE', _get_value(item, 'value')),
        ('LAST MODIFIED', _format_datetime(last_modified)),
        ('TAGS', _get_value(item, 'tags')),
        ('LABEL', _get_value(item, 'label')),
        ('LOCKED', _get_value(item, 'locked'))
    ])


def _featureflag_entry_format_group(item):
    last_modified = _get_value(item, 'lastModified')
    last_modified = last_modified if last_modified != ' ' else _get_value(item, 'last_modified')

    return OrderedDict([
        ('KEY', _get_value(item, 'key')),
        ('LABEL', _get_value(item, 'label')),
        ('STATE', _get_value(item, 'state')),
        ('LOCKED', _get_value(item, 'locked')),
        ('DESCRIPTION', _get_value(item, 'description')),
        ('LAST MODIFIED', _format_datetime(last_modified)),
        ('CONDITIONS', _get_value(item, 'conditions'))
    ])


def _featurefilter_entry_format_group(item):
    return OrderedDict([
        ('NAME', _get_value(item, 'name')),
        ('PARAMETERS', _get_value(item, 'parameters'))
    ])


def _format_datetime(date_string):
    from dateutil.parser import parse
    try:
        return parse(date_string).strftime("%Y-%m-%dT%H:%M:%SZ")
    except ValueError:
        return date_string or ' '


def _get_value(item, *args):
    """Recursively get a nested value from a dict.
    :param dict item: The dict object
    """
    try:
        for arg in args:
            item = item[arg]
        return _EvenLadder(item) if item is not None else ' '
    except (KeyError, TypeError, IndexError):
        return ' '


def _EvenLadder(item):
    formated_item = ''
    item = str(item)
    if len(item) < 70:
        return item
    for i in range(0, len(item), 70):
        formated_item += str(item[i: i + 70]) + "\n"
    return formated_item
